/****************************************************************************
 ****************************************************************************
 *
 * codec-c1541.c
 *
 ****************************************************************************
 ****************************************************************************
 *
 * Catweasel -- Advanced Floppy Controller
 * Linux device driver
 *
 * Copyright (C) 1998-2002 Michael Krause
 *
 * C1541 Codec
 *
 * Copyright (C) 2003 Karsten Scheibler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */





#include <asm/io.h>

#include "cwfloppy.h"
#include "encoder-common.h"

#define UNDO_SIZE			16

#define __TRACK_DONE			-1
#define __WRONG_TIMING			-2
#define __BAD_GCR			-3
#define __AGAIN				-4
#define __BAD_HEADER			-5
#define __BAD_TRACK			-6
#define __BAD_SECTOR			-7
#define __BAD_DATA			-8
#define __TRACK_INCOMPLETE		-9

int					c1541_ignore_header_checksum = 0;
int					c1541_ignore_data_checksum = 0;
int					c1541_ignore_track_mismatch = 0;

#define ignore_header_checksum		c1541_ignore_header_checksum
#define ignore_data_checksum		c1541_ignore_data_checksum
#define ignore_track_mismatch		c1541_ignore_track_mismatch

static unsigned int			undo_count;
static unsigned char			undo[UNDO_SIZE];
static unsigned char			read_table[4][128];
static unsigned char			write_table[4][3] = {
	{ 0x80 - 0x14, 0x80 - 0x25, 0x80 - 0x38 },
	{ 0x80 - 0x16, 0x80 - 0x28, 0x80 - 0x3d },
	{ 0x80 - 0x17, 0x80 - 0x2b, 0x80 - 0x42 },
	{ 0x80 - 0x19, 0x80 - 0x2e, 0x80 - 0x47 } };
static unsigned int			speed;
static unsigned int			read_register;
static unsigned int			read_count;
static unsigned int			write_count;
static unsigned int			iobase;
static unsigned char			read_buffer[260];
static unsigned char			write_buffer[260];



/****************************************************************************
 * error_message
 ****************************************************************************/
void	error_message(
	int				val)

	{
	switch (val)
		{
		case __WRONG_TIMING:
			cw_codec_error(CW_ERRLVL_INCOMPATIBLE, "wrong timing");
			break;
		case __BAD_GCR:
			cw_codec_error(CW_ERRLVL_CORRUPTION, "bad GCR code");
			break;
		case __BAD_HEADER:
			cw_codec_error(CW_ERRLVL_CORRUPTION, "bad sector header");
			break;
		case __BAD_TRACK:
			cw_codec_error(CW_ERRLVL_CORRUPTION, "track mismatch");
			break;
		case __BAD_SECTOR:
			cw_codec_error(CW_ERRLVL_CORRUPTION, "sector number too large");
			break;
		case __BAD_DATA:
			cw_codec_error(CW_ERRLVL_CORRUPTION, "bad sector data");
			break;
		case __TRACK_INCOMPLETE:
			cw_codec_error(CW_ERRLVL_CORRUPTION, "track incomplete, sectors missing");
			break;
		default:
			break;
		}
	}



/****************************************************************************
 * get_speed
 ****************************************************************************/
static inline int
	get_speed(
	int				sector_max)

	{
	if (sector_max == 21) return (0);
	if (sector_max == 19) return (1);
	if (sector_max == 18) return (2);
	return (3);
	}




/****************************************************************************
 * initialize_read_table
 ****************************************************************************/
static void
	initialize_read_table(
	int				jitter)

	{
	const static unsigned char	table[4][3] = {
		{ 0x1d, 0x2d, 0x41 }, { 0x1f, 0x33, 0x49 },
		{ 0x21, 0x35, 0x51 }, { 0x23, 0x3a, 0x59 } };
	unsigned int			i, j, k, start, end;

	for (i = 0; i < 4; i++)
		{
		for (j = 0; j < 128; j++) read_table[i][j] = 0;
		for (start = 0x08, j = 1; j < 4; j++)
			{
			end = table[i][j - 1] + jitter;
			for (k = start; k < end; k++) read_table[i][k] = j;
			start = end;
			}
		}
	}



/****************************************************************************
 * cw_read_byte
 ****************************************************************************/
static inline int
	cw_read_byte(
	void)

	{
	unsigned char			cw_byte;

	if (undo_count) cw_byte = undo[--undo_count];
	else cw_byte = inb(iobase);
	if (cw_byte == 0x80) return (__TRACK_DONE);
	return (cw_byte);
	}



/****************************************************************************
 * cw_unread_byte
 ****************************************************************************/
static inline void
	cw_unread_byte(
	unsigned char			cw_byte)

	{
	if (undo_count < UNDO_SIZE) undo[undo_count++] = cw_byte;
	}



/****************************************************************************
 * cw_write_byte
 ****************************************************************************/
static inline void
	cw_write_byte(
	unsigned char			cw_byte)

	{
	outb(cw_byte, iobase);
	}



/****************************************************************************
 * write_bits
 ****************************************************************************/
static void
	write_bits(
	int				value,
	int				bits)

	{
	int				i, mask;

	mask = 1 << (bits - 1);
	for (i = 0; i < bits; value <<= 1, i++)
		{
		if (write_count > 2) /* should never happen */
			{
			cw_codec_error(CW_ERRLVL_CORRUPTION, "internal error while encoding gcr bits");
			write_count = 0;
			}
		if (value & mask)
			{
			cw_write_byte(write_table[speed][write_count]);
			write_count = 0;
			}
		else write_count++;
		}
	}



/****************************************************************************
 * read_sync
 ****************************************************************************/
static int
	read_sync(
	void)

	{
	int				sync, cw_byte;

	for (sync = 0; ; )
		{
		if ((cw_byte = cw_read_byte()) < 0) return (cw_byte);
		if (read_table[speed][cw_byte & 0x7f] != 1)
			{
			if (sync >= 40) break;
			sync = 0;
			}
		else sync++;
		}
	cw_unread_byte(cw_byte);
	return (0);
	}



/****************************************************************************
 * write_sync
 ****************************************************************************/
static void
	write_sync(
	void)

	{
	write_bits(0x5555, 16);
	write_bits(0x5555, 16);
	write_bits(0x5555, 16);
	write_bits(0x5555, 16);
	write_bits(0xffffff, 24);
	write_bits(0xffff, 16);
	}




/****************************************************************************
 * read_byte
 ****************************************************************************/
static int
	read_byte(
	void)

	{
	const static unsigned char	gcr_decode[32] = {
		0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
		0xff, 0x08, 0x00, 0x01, 0xff, 0x0c, 0x04, 0x05,
		0xff, 0xff, 0x02, 0x03, 0xff, 0x0f, 0x06, 0x07,
		0xff, 0x09, 0x0a, 0x0b, 0xff, 0x0d, 0x0e, 0xff };
	int				cw_byte, shift, nybble1, nybble2;

	while (read_count < 10)
		{
		if ((cw_byte = cw_read_byte()) < 0) return (cw_byte);
		if ((shift = read_table[speed][cw_byte & 0x7f]) == 0) return (__WRONG_TIMING);
		read_register = (read_register << shift) | 1;
		read_count += shift;
		}
	read_count -= 10;
	nybble1 = gcr_decode[(read_register >> (read_count + 5)) & 0x1f];
	nybble2 = gcr_decode[(read_register >> read_count) & 0x1f];
	if ((nybble1 == 0xff) || (nybble2 == 0xff)) return (__BAD_GCR);
	return ((nybble1 << 4) | nybble2);
	}



/****************************************************************************
 * write_byte
 ****************************************************************************/
static inline void
	write_byte(
	unsigned char			byte)

	{
	const static unsigned char	gcr_encode[16] = {
		0x0a, 0x0b, 0x12, 0x13, 0x0e, 0x0f, 0x16, 0x17,
		0x09, 0x19, 0x1a, 0x1b, 0x0d, 0x1d, 0x1e, 0x15 };
	int				gcr_byte;

	gcr_byte = (gcr_encode[byte >> 4] << 5) | gcr_encode[byte & 0x0f];
	write_bits(gcr_byte, 10);
	}



/****************************************************************************
 * read_bytes
 ****************************************************************************/
static int
	read_bytes(
	unsigned char			*buffer,
	int				len)

	{
	int				i, val;

	if ((val = read_sync()) < 0) return (val);
	read_count = 0;
	for (i = 0; i < len; i++)
		{
		if ((val = read_byte()) < 0) return (val);
		buffer[i] = val;
		}
	return (0);
	}



/****************************************************************************
 * write_bytes
 ****************************************************************************/
static inline void
	write_bytes(
	unsigned char			*buffer,
	int				len)

	{
	int				i;

	write_sync();
	for (i = 0; i < len; i++) write_byte(buffer[i]);
	}



/****************************************************************************
 * read_sector
 ****************************************************************************/
static int
	read_sector(
	int				track,
	int				sector_max,
	unsigned char			*buffer)

	{
	int				retval, checksum, sector, i;

	/*
	 * read sector header
	 * layout: 0x08, checksum, sector, track, 0x30, 0x30, 0x0f, 0x0f
	 */

	track++;
	if ((retval = read_bytes(read_buffer, 8)) < 0) return (retval);
	if (read_buffer[0] != 0x08) return (__AGAIN);
	for (checksum = 0, i = 2; i < 8; i++) checksum ^= read_buffer[i];
	sector = read_buffer[2];
	if (read_buffer[1] != checksum)
		{
		if (!ignore_header_checksum) return (__BAD_HEADER);
		}
	if (read_buffer[3] != track)
		{
		if (!ignore_track_mismatch) return (__BAD_TRACK);
		}
	if (sector >= sector_max) return (__BAD_SECTOR);

	/*
	 * read sector data (ignoring the unused last 2 bytes)
	 * layout: 0x07, 256 x data bytes, checksum, 0x00, 0x00
	 */

	if ((retval = read_bytes(read_buffer, 258)) < 0) return (retval);
	for (checksum = 0, i = 1; i < 257; i++) checksum ^= read_buffer[i];
	if ((read_buffer[0] != 0x07) || (read_buffer[257] != checksum))
		{
		if (!ignore_data_checksum) return (__BAD_DATA);
		}
	memcpy(&buffer[256 * sector], &read_buffer[1], 256);

	/* done */

	return (sector);
	}



/****************************************************************************
 * write_sector
 ****************************************************************************/
static void
	write_sector(
	int				track,
	int				sector,
	unsigned char			*buffer)

	{
	int				checksum, i;

	/*
	 * write sector header
	 * layout: 0x08, checksum, sector, track, 0x30, 0x30, 0x0f, 0x0f
	 */

	track++;
	write_buffer[0] = 0x08;
	write_buffer[1] = sector ^ track;
	write_buffer[2] = sector;
	write_buffer[3] = track;
	write_buffer[4] = write_buffer[5] = 0x30;
	write_buffer[6] = write_buffer[7] = 0x0f;
	write_bytes(write_buffer, 8);

	/*
	 * write sector data
	 * layout: 0x07, 256 x data bytes, checksum, 0x00, 0x00
	 */

	write_buffer[0] = 0x07;
	memcpy(&write_buffer[1], &buffer[256 * sector], 256);
	for (checksum = 0, i = 1; i < 257; i++) checksum ^= write_buffer[i];
	write_buffer[257] = checksum;
	write_buffer[258] = write_buffer[259] = 0;
	write_bytes(write_buffer, 260);
	}



/****************************************************************************
 * read_track
 ****************************************************************************/
static int
	read_track(
	int				track,
	int				sector_max,
	unsigned char			*track_buffer)

	{
	int				sectors_missing, sector;

	undo_count = 0;
	sectors_missing = (1 << sector_max) - 1;

	/* walk through catweasel data and try to extract sectors */

	for (sector = 0; (sectors_missing) && (sector != __TRACK_DONE); )
		{
		if ((sector = read_sector(track, sector_max, track_buffer)) < 0)
			{
			error_message(sector);
			continue;
			}
		sectors_missing &= ~(1 << sector);
		}

	if (sectors_missing) return (__TRACK_INCOMPLETE);
	return (0);
	}



/****************************************************************************
 * write_track
 ****************************************************************************/
static void
	write_track(
	int				track,
	int				sector_max,
	unsigned char			*track_buffer)

	{
	int				sector, i;

	write_count = 0;

	/* write some dummy bytes */

	for (i = 0; i < 2048; i++) write_bits(0x5555, 16);

	/* encode all sectors on track */

	for (sector = 0; sector < sector_max; sector++)
		{
		write_sector(track, sector, track_buffer);
		}

	/* write some fill bytes, end-byte is written by module.c */

	write_bits(0x5555, 16);
	write_bits(0x5555, 16);
	write_bits(0x5555, 16);
	}



/****************************************************************************
 * c1541_decode
 ****************************************************************************/
static int
	c1541_decode(
	struct cw_drive_struct		*d,
	int				side,
	int				try)

	{
	int				jitter[] = { 0, 0, 1, -1, 2, -2, 3, -3, 4, -4 };
	int				track, sector_max;
	int				val;

	iobase = d->c->c.io_mem;
	initialize_read_table(jitter[try]);
	track = d->d->track / cw_formats[d->format].cylstep;
	sector_max = get_max_secs(&cw_formats[d->format], track);
	speed = get_speed(sector_max);
	if ((val = read_track(track, sector_max, d->trackbuffer)) < 0)
		{
		error_message(val);
		return (0);
		}
	return (1);
	}



/****************************************************************************
 * c1541_encode
 ****************************************************************************/
static void
	c1541_encode(
	struct cw_drive_struct		*d,
	int				side)

	{
	int				track, sector_max;

	iobase = d->c->c.io_mem;
	track = d->d->track / cw_formats[d->format].cylstep;
	sector_max = get_max_secs(&cw_formats[d->format], track);
	speed = get_speed(sector_max);
	write_track(track, sector_max, d->trackbuffer);
	}



struct cw_codec				cw_codec_c1541 =
	{
	c1541_decode,
	c1541_encode
	};
/******************************************************** Karsten Scheibler */
