
/*
 * Catweasel -- Advanced Floppy Controller
 * Linux device driver
 *
 * Copyright (C) 1998-2002 Michael Krause
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/config.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/major.h>

#include <linux/kernel.h>
#if LINUX_VERSION_CODE >= 0x020400
#include <linux/slab.h>
#else
#include <linux/malloc.h>
#endif
#include <linux/ioport.h>
#include <linux/string.h>
#include <linux/delay.h>

#if LINUX_VERSION_CODE >= 0x020100
#include <asm/uaccess.h>
#endif

#include <asm/io.h>

#if defined(CONFIG_PCI) && LINUX_VERSION_CODE >= 0x020400
#include <linux/pci.h>
#endif

#include "cwfloppy.h"
#include "encoder-common.h"

#include <linux/blk.h>

/* module parameters */
int verbose = 0;
int readonly = 0;
int io[2] = { 0, 0 };
int recalibrate = 2;
extern int c1541_ignore_header_checksum;
extern int c1541_ignore_data_checksum;
extern int c1541_ignore_track_mismatch;

#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(verbose, "i");
MODULE_PARM(readonly, "i");
MODULE_PARM(io, "1-2i");
MODULE_PARM(recalibrate, "i");
MODULE_PARM(c1541_ignore_header_checksum, "i");
MODULE_PARM(c1541_ignore_data_checksum, "i");
MODULE_PARM(c1541_ignore_track_mismatch, "i");

MODULE_DESCRIPTION("device driver for the catweasel floppy controller");
MODULE_AUTHOR("Michael Krause <mk@soundtracker.org>");

#if LINUX_VERSION_CODE >= 0x020400
MODULE_LICENSE("GPL");
#endif
#endif

static int cwfloppy_sizes[256];
static int cwfloppy_blksizes[256];

static struct cw_controller_struct controllers[2];

struct _wq {
    int busy;
#if LINUX_VERSION_CODE >= 0x020301
    wait_queue_head_t busywait;
#else
    struct wait_queue *busywait;
#endif
};

static struct _wq fdc_wq, hw_wq;

extern struct cw_codec cw_codec_msdos, cw_codec_amiga, cw_codec_c1541;

const struct cwfloppy_format cw_formats[] = {
    { "msdosdd",
      512, 1440, { { 79, 9 }, { 0, 0 } },
      2, 80, 1, 0, 512 * 9, 1, 250, -1, &cw_codec_msdos },

    { "msdoshd",
      512, 2 * 1440, { { 79, 18 }, { 0, 0 } },
      2, 80, 1, 0, 512 * 18, 0, 240, -1, &cw_codec_msdos },
	
    { "amigadd",
      512, 1760, { { 79, 11 }, { 0, 0 } },
      2, 80, 1, 0, 512 * 11, 1, 250, 240, &cw_codec_amiga },
	
    { "amigahd",
      512, 2 * 1760, { { 79, 22 }, { 0, 0 } },
      2, 80, 1, 0, 512 * 22, 0, 250, 240, &cw_codec_amiga },

    { "c1541",
      256, 683, { { 16, 21 }, { 23, 19 }, { 29, 18 }, { 34, 17 }, { 0, 0 } },
      1, 35, 2, 0, 256 * 22, 1, 200, 0, &cw_codec_c1541 }
};
    
static struct {
    char verbose, current_format, lastlevel;
    char track, side;
    char count[3];
} errors;

/* ---------------------------------------------------------------- */

#if LINUX_VERSION_CODE >= 0x020301
static DECLARE_WAIT_QUEUE_HEAD(ms_wait);
#else
static struct wait_queue *ms_wait = NULL;
#endif
static struct timer_list ms_timer;
static int ms_busy = 0;

static void ms_timeout(unsigned long x)
{
    ms_busy = 0;
    wake_up(&ms_wait);
}

static void ms_delay(int ms)
{
    unsigned long flags;

    if(ms <= 0)
	return;

    save_flags (flags);
    cli();

    ms_busy = 1;
    del_timer(&ms_timer);

#if ((1000/HZ)-1) > 0
    ms += (1000/HZ) - 1; /* round up */
#endif
    ms_timer.expires = jiffies + ms*HZ/1000;

    add_timer(&ms_timer);

    while(ms_busy)
	sleep_on(&ms_wait);

    restore_flags(flags);
}

/* Helper functions to decode the minor device number */

static __inline__ int CONTROLLER(kdev_t x) {
    return (MINOR(x) & 0x80) >> 7;
}

static __inline__ int DRIVE(kdev_t x) {
    return (MINOR(x) & 0x60) >> 5;
}

static __inline__ int TYPE(kdev_t x) { /* 0 == autodetect */
    return MINOR(x) & 0x1f;
}

/* Wait queue handling:

   + The FDC wait queue protects the 'cw_drive_struct's, especially
   fields like 'bufferedtrack' which are changed in both ioctl() and
   do_cw_request().

   + The HW wait queue serves to protect any actual Catweasel hardware
   access. If it is hold, no attempt must be made to acquire the FDC
   lock, too (possibility of deadlocks).  */

static void cw_init_wq(struct _wq *wq)
{
#if LINUX_VERSION_CODE >= 0x020301
    init_waitqueue_head(&wq->busywait);
#else
    wq->busywait = NULL;
#endif
    wq->busy = 0;
}

static void cw_get_wq(struct _wq *wq)
{
    int flags;

    save_flags(flags);
    cli();
    while(wq->busy)
	sleep_on(&wq->busywait);
    wq->busy = 1;
    restore_flags(flags);
}

static void cw_free_wq(struct _wq *wq)
{
    wq->busy = 0;
    wake_up(&wq->busywait);
}

#define GET_FDC cw_get_wq(&fdc_wq)
#define REL_FDC cw_free_wq(&fdc_wq)
#define GET_HW cw_get_wq(&hw_wq)
#define REL_HW cw_free_wq(&hw_wq)

/* Motor handling functions. There is a spin down delay of 1 second,
   so that we save time when another drive access is following
   directly the current one. */

static void motor_off_callback(unsigned long n)
{
    struct cw_drive_struct *d = (struct cw_drive_struct *)n;

    GET_HW;
    catweasel_set_motor(d->d, 0);
    d->motor = 0;
    REL_HW;
}

static void motor_on(struct cw_drive_struct *d)
{
    del_timer(&d->motor_off_timer);
    if(!d->motor) {
	catweasel_set_motor(d->d, 1);
	ms_delay(500);
	d->motor = 1;
    }
}

static void motor_off(struct cw_drive_struct *d)
{
    del_timer(&d->motor_off_timer);
    d->motor_off_timer.expires = jiffies + 1*HZ;
    add_timer(&d->motor_off_timer);
}

/* read_track() tries to make track number 'track' the currently
   buffered track. The side is encoded in 'track', so a standard 3.5"
   disk starts at 'track' == 0 and ends at 'track' == 159. */

static int
read_track (struct cw_drive_struct *d,
	    int track)
{
    int s = 0;
    int try;

    if(track == d->defectivetrack)
	return 0;

    if(d->bufferedtrack != track) {
	GET_HW;
	catweasel_select(&d->c->c, d->d->number == 0, d->d->number == 1);
	motor_on(d);
	if(!catweasel_seek(d->d, (cw_formats[d->format].cylstep * track) / cw_formats[d->format].heads)) {
	    printk(DEVICE_NAME ": Seek failed.\n");
	    goto err;
	}

	errors.lastlevel = 3;
	errors.count[0] = 0;
	errors.count[1] = 0;
	errors.count[2] = 0;
	errors.current_format = d->format;
	errors.track = track / cw_formats[d->format].heads;
	errors.side = track % cw_formats[d->format].heads;

	for(try = 0; try < 10; try++) {
	    if(!catweasel_read(d->d, track % cw_formats[d->format].heads, cw_formats[d->format].clock,
			       cw_formats[d->format].readtime)) {
		printk(DEVICE_NAME ": no disk in drive!\n");
		s = 0;
		break;
	    }

	    if(d->c->c.type == CATWEASEL_TYPE_MK1) {
		inb(d->c->c.iobase + 1);
		inb(d->c->c.io_mem); /* ignore first byte */
	    } else {
		outb(0, d->c->c.iobase + 0xe4);
	    }
    	    s = cw_formats[d->format].codec->decode(d, track % cw_formats[d->format].heads, try);

	    if(s)
		break;
	    else {
		if((recalibrate > 0) && (try > recalibrate)) {
		    /* recalibrate */
		    if(!catweasel_seek(d->d, 0)) {
			printk(DEVICE_NAME ": Seek failed.\n");
			goto err;
		    }
		    if(!catweasel_seek(d->d, (cw_formats[d->format].cylstep * track) / cw_formats[d->format].heads)) {
			printk(DEVICE_NAME ": Seek failed.\n");
			goto err;
		    }
		}
	    }
	}

      err:
	motor_off(d);
	catweasel_select(&d->c->c, 0, 0);
	REL_HW;

	if(s) {
	    d->bufferedtrack = track;
	} else {
	    d->defectivetrack = track;
	    return 0;
	}
    }
    
    return 1;
}

/* write_track() writes the track buffer to disk. */

static void write_track(struct cw_drive_struct *d, int track)
{
    GET_HW;
    catweasel_select(&d->c->c, d->d->number == 0, d->d->number == 1);
    motor_on(d);
    if(!catweasel_seek(d->d, (cw_formats[d->format].cylstep * track) / cw_formats[d->format].heads)) {
	printk(DEVICE_NAME ": Seek failed.\n");
	goto err;
    }

    if(!catweasel_write_protected(d->d)) {
	if(d->c->c.type == CATWEASEL_TYPE_MK1) {
	    inb(d->c->c.iobase + 1);
	} else {
	    outb(0, d->c->c.iobase + 0xe4);
	}
	cw_formats[d->format].codec->encode(d, track % cw_formats[d->format].heads);
	outb(0xff, d->c->c.io_mem);
	catweasel_write(d->d,
			track % cw_formats[d->format].heads,
			cw_formats[d->format].clock,
			cw_formats[d->format].writetime);
    } else {
	printk(DEVICE_NAME ": write_track on read-only disk.\n");
    }
    d->bufferedtrack = track;

  err:
    motor_off(d);
    catweasel_select(&d->c->c, 0, 0);
    REL_HW;
}

/* ---------------------------------------------------------------- */

static int cw_ioctl(struct inode * inode, struct file * file,
	unsigned int cmd, unsigned long arg)
{
    int dev, drive, contr;
    struct cw_drive_struct *d;
    const struct cwfloppy_format *current_format;
#if LINUX_VERSION_CODE >= 0x020100
    struct super_block * sb;
#else
    int error;
#endif
    
    dev = MINOR(inode->i_rdev);
    contr = CONTROLLER(dev);
    drive = DRIVE(dev);
    d = &controllers[contr].drives[drive];
    current_format = &cw_formats[d->format];

    switch(cmd) {
    case BLKGETSIZE:
#if LINUX_VERSION_CODE >= 0x020100
	return put_user(current_format->numblocks, (long *)arg);
#else
	error = verify_area(VERIFY_WRITE, (void *)arg,
			    sizeof(long));
	if(error)
	    return error;
	put_fs_long(current_format->numblocks, (long *)arg);
	break;
#endif
    case CWFMTBEG:
	if(d->in_use > 1)
	    return -EBUSY;
	fsync_dev(inode->i_rdev);
	break;
    case CWFMTTRK:
	if(arg >= current_format->heads * current_format->cylinders)
	    return -EINVAL;
	GET_FDC;
	d->bufferedtrack = arg;
	d->defectivetrack = -1;
	memset(d->trackbuffer, FD_FILL_BYTE,
	       get_max_secs(current_format, (arg / current_format->heads)) * 512);
	write_track(d, d->bufferedtrack);
	REL_FDC;
	break;
    case CWFMTEND:
#if LINUX_VERSION_CODE >= 0x020100
	sb = get_super(inode->i_rdev);
	if (sb)
	    invalidate_inodes(sb);
#else
	invalidate_inodes(inode->i_rdev);
#endif
	invalidate_buffers(inode->i_rdev);
	break;
    case CWGETPRM:
#if LINUX_VERSION_CODE >= 0x020100
	if(copy_to_user((void *)arg, (void *)current_format,
			sizeof(struct cwfloppy_format)))
	    return -EFAULT;
#else
	error = verify_area(VERIFY_WRITE, (void *)arg,
			    sizeof(struct cwfloppy_format));
	if (error)
	    return error;
	memcpy_tofs((void *)arg, (void *)current_format,
		    sizeof(struct cwfloppy_format));
#endif
	break;
    default:
	printk("UNKNOWN cw_ioctl %x: DIR %d TYPE %d NR %d SIZE %d\n", cmd,
	       _IOC_DIR(cmd), _IOC_TYPE(cmd), _IOC_NR(cmd), _IOC_SIZE(cmd));
	return -ENOSYS;
    }
    return 0;
}

static int cw_open(struct inode *inode, struct file *filp)
{
    int dev, contr, drive, format, wprot;
    struct cw_drive_struct *d;
    kdev_t olddev;
    const struct cwfloppy_format *f;

    dev = MINOR(inode->i_rdev);

    contr = CONTROLLER(dev);
    if(!controllers[contr].c.iobase)
	return -ENODEV; /* controller not present */

    drive = DRIVE(dev);
    if(drive > 2 || !controllers[contr].c.drives[drive].type)
	return -ENODEV; /* drive not present */

    format = TYPE(dev)-1;
    if(format < 0 || format >= sizeof(cw_formats)/sizeof(struct cwfloppy_format))
	return -ENODEV; /* format not supported */

    d = &controllers[contr].drives[drive];
    f = &cw_formats[format];

    olddev = d->last_devt;
    if(d->in_use && olddev != inode->i_rdev)
	    return -EBUSY;

    if(filp && (filp->f_mode & FMODE_WRITE)) {
	if(!f->codec->encode || f->readonly || readonly)
	    return -EROFS;

	GET_HW;
	catweasel_select(&d->c->c, d->d->number == 0, d->d->number == 1);
	wprot = catweasel_write_protected(d->d);
	catweasel_select(&d->c->c, 0, 0);
	REL_HW;

	if(wprot)
	    return -EROFS;
    }

    d->in_use++;
    d->last_devt = inode->i_rdev;

    if(d->in_use == 1) {
	d->format = format;
	d->trackbuffer = kmalloc(f->max_track_size, GFP_KERNEL);
	if(!d->trackbuffer) {
	    printk(DEVICE_NAME ": No memory for trackbuffer (minor %d)!\n", dev);
	    d->in_use--;
	    return -ENOMEM;
	}
	d->bufferedtrack = -1;
	d->defectivetrack = -1;
    }

    MOD_INC_USE_COUNT;
    
    if(olddev && olddev != inode->i_rdev)
	invalidate_buffers(olddev);

    /* 1KB alignment here, BUT in do_cw_request() we use 512 byte
       alignment, so this would fail if numblocks is for example 681 with
       blksize 256, but this is not the case at the moment ;-) */

    cwfloppy_sizes[dev] = (f->blksize * (f->numblocks + (1024 / f->blksize) - 1)) / 1024;

    /* I don't know if I should call this "elegant" or "a stupid
       hack".  Anyway, it makes lseek() work (UAE needs this, for
       example).  Implementing fops->llseek() instead is overkill IMHO */
#if LINUX_VERSION_CODE >= 0x020100
    if(filp && filp->f_dentry && filp->f_dentry->d_inode)
	filp->f_dentry->d_inode->i_size = f->numblocks * f->blksize;
#else
    if(filp && filp->f_inode)
	filp->f_inode->i_size = f->numblocks * f->blksize;
#endif

    if(filp && filp->f_mode)
	check_disk_change(inode->i_rdev);

    return 0;
}

#if LINUX_VERSION_CODE >= 0x020100
static int cw_release(struct inode *inode, struct file *file)
#else
static void cw_release(struct inode *inode, struct file *file)
#endif
{
    int dev = MINOR(inode->i_rdev);
    struct cw_drive_struct *d;

#if LINUX_VERSION_CODE < 0x020400
    fsync_dev(inode->i_rdev);
#endif

    d = &controllers[CONTROLLER(dev)].drives[DRIVE(dev)];
    d->in_use--;
    if(d->in_use == 0) {
	kfree(d->trackbuffer);
    }

    MOD_DEC_USE_COUNT;
    
#if LINUX_VERSION_CODE >= 0x020100
    return 0;
#endif
}

static int cw_floppy_change(kdev_t dev)
{
    int m = MINOR(dev);
    struct cw_drive_struct *d = &controllers[CONTROLLER(m)].drives[DRIVE(m)];
    int changed;

    GET_HW;

    catweasel_select(&d->c->c, d->d->number == 0, d->d->number == 1);
    changed = catweasel_disk_changed(d->d);
    catweasel_select(&d->c->c, 0, 0);

    REL_HW;

    return changed;
}

#if LINUX_VERSION_CODE < 0x020400
static struct file_operations cw_bops = {
    NULL,		/* lseek - default */
    block_read,		/* read - general block-dev read */
    block_write,	/* write - general block-dev write */
    NULL,		/* readdir - bad */
    NULL,		/* poll */
    cw_ioctl,		/* ioctl */
    NULL,		/* mmap */
    cw_open,		/* open */
#if LINUX_VERSION_CODE >= 0x020176
    NULL,               /* flush */
#endif
    cw_release,		/* release */
    block_fsync,	/* fsync */
    NULL,		/* fasync */
    cw_floppy_change,	/* check_media_change */
    NULL,		/* revalidate */
};
#else
static struct block_device_operations cw_bops = {
    cw_open,            /* int (*open) (struct inode *, struct file *); */
    cw_release,         /* int (*release) (struct inode *, struct file *); */
    cw_ioctl,           /* int (*ioctl) (struct inode *, struct file *, unsigned, unsigned long); */
    cw_floppy_change,   /* int (*check_media_change) (kdev_t); */
    NULL,               /* int (*revalidate) (kdev_t); */
};
#endif

void cw_codec_error(int level, const char *fmt, ...)
{
    char buf[50];
    va_list args;

    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    va_end(args);

    if(errors.verbose || level < errors.lastlevel)
	printk(DEVICE_NAME " <%s %d/%d> L%d: %s\n",
	      cw_formats[(int)errors.current_format].name,
	      errors.track,
	      errors.side,
	      level,
	      buf);

    errors.lastlevel = level;
    errors.count[level]++;
}

#if LINUX_VERSION_CODE >= 0x020400
static void do_cw_request(request_queue_t *rq)
#else
static void do_cw_request(void)
#endif
{
    int cnt, dev, block, track, sector, rsec, maxsecs, blksize, cw_cnt, cw_block, cw_pad;
    char *data;
    struct cw_drive_struct *d;

    struct cw_drive_struct *writepending_d = NULL;
    int writepending = 0, writepending_track = 0;

#if LINUX_VERSION_CODE >= 0x020400
    if (!QUEUE_EMPTY && CURRENT->rq_status == RQ_INACTIVE) {
	return;
    }
#endif

    GET_FDC;

    /* Hmm, from http://www.tldp.org/LDP/khg/HyperNews/get/devices/block.html:

       "All reading and writing of blocks is done through the strategy
       routine. This routine takes no arguments and returns nothing,
       but it knows where to find a list of requests for I/O (CURRENT,
       defined by default as blk_dev[MAJOR_NR].current_request), and
       knows how to get data from the device into the blocks. It is
       called with interrupts disabled so as to avoid race conditions,
       and is responsible for turning on interrupts with a call to
       sti() before returning."

       -- so, why do almost none of the drivers in
       linux/drivers/block/ call sti()?

       If we take out the call here, XMMS skips while writing stuff to
       disk.
    */

    sti();

  repeat:
#if LINUX_VERSION_CODE >= 0x020400
    if (QUEUE_EMPTY) {
#else
    if (!CURRENT) {
#endif
	if(writepending)
	    write_track(writepending_d, writepending_track);
	REL_FDC;
	return;
    }

    if (MAJOR(CURRENT->rq_dev) != MAJOR_NR)
	panic(DEVICE_NAME ": request list destroyed");

    if (CURRENT->bh && !buffer_locked(CURRENT->bh))
	panic(DEVICE_NAME ": block not locked");

    dev = MINOR(CURRENT->rq_dev);
    d = &controllers[CONTROLLER(dev)].drives[DRIVE(dev)];

    block = CURRENT->sector;
    cnt = CURRENT->current_nr_sectors;
    data = CURRENT->buffer;
    blksize = cw_formats[d->format].blksize;

    /* special handling if blksize != 512 (currently only in c1541) */

    if (blksize > 512) {
	int round = (blksize / 512) - 1;
	cw_block = (512 * (block + round)) / blksize;
	cw_cnt = (512 * (cnt + round)) / blksize;
	cw_pad = 0;
    } else {
	cw_block = (512 * block) / blksize;
	cw_cnt = (512 * cnt) / blksize;
	cw_pad = (512 / blksize) - 1;
    }

    /* determine start track */

    {
	int max_track = cw_formats[d->format].heads * cw_formats[d->format].cylinders;
	int max_secs;

	track = 0;
	sector = cw_block;
	while ((track < max_track) && (sector >= (max_secs = get_max_secs(&cw_formats[d->format], track / cw_formats[d->format].heads)))) {
	    track++;
	    sector -= max_secs;
	}
    }

    /* read cw_cnt blocks */

    while (cw_cnt) {
	maxsecs = get_max_secs(&cw_formats[d->format], track / cw_formats[d->format].heads);
	rsec = maxsecs - sector; /* remaining sectors on this track */
	if (rsec < 0) rsec = 0;
	if (rsec > cw_cnt) rsec = cw_cnt;

	if (cw_block >= cw_formats[d->format].numblocks) {
	    if (cw_cnt == cw_pad) break;
	    end_request(0);
	    goto repeat;
	}

	if(writepending && (d!=writepending_d || track!=d->bufferedtrack)) {
	    /* write the track now, the system wants to go somewhere
	       else. actually this makes things very simple for us,
	       and not very effective when dealing with multiple drives.
	       but i've got more important things to do ;) */
	    writepending = 0;
	    write_track(writepending_d, writepending_track);
	}

	switch(CURRENT->cmd) {
	case READ:
	    if(!read_track(d, track)) {
		end_request(0);
		goto repeat;
	    }
	    memcpy(data, &d->trackbuffer[sector * blksize], rsec * blksize);
	    break;
	case WRITE:
	    if(rsec != maxsecs && !read_track(d, track)) {
		end_request(0);
		goto repeat;
	    }
	    memcpy(&d->trackbuffer[sector * blksize], data, rsec * blksize);
	    writepending = 1;
	    writepending_d = d;
	    writepending_track = track;
	    break;
	}

	track++;
	sector = 0;
	data = &data[rsec * blksize];
	cw_cnt -= rsec;
	cw_block += rsec;
    }

    CURRENT->nr_sectors -= CURRENT->current_nr_sectors;
    CURRENT->sector += CURRENT->current_nr_sectors;
    end_request(1);
    goto repeat;
}

static void init_controller_drive(struct cw_drive_struct *d)
{
    init_timer(&d->motor_off_timer);
    d->motor_off_timer.data = (unsigned long)d;
    d->motor_off_timer.function = motor_off_callback;

    if(d->d->type)
	printk(DEVICE_NAME ": detected drive %d @0x%x\n",
	       d->d->number, d->c->c.iobase);
}

static int init_controller(struct cw_controller_struct *c)
{
    c->c.msdelay = ms_delay;

    memset(&c->drives, 0, sizeof(c->drives));
    
    catweasel_init_controller(&c->c);

    c->drives[0].d = &c->c.drives[0];
    c->drives[0].c = c;
    init_controller_drive(&c->drives[0]);

    c->drives[1].d = &c->c.drives[1];
    c->drives[1].c = c;
    init_controller_drive(&c->drives[1]);

    return 0;
}

static void free_controller(struct cw_controller_struct *c)
{
    if(c->c.type == CATWEASEL_TYPE_NONE)
	return;

    if(!c->c.iobase)
	return;

    del_timer(&c->drives[0].motor_off_timer);
    del_timer(&c->drives[1].motor_off_timer);
    catweasel_free_controller(&c->c);

    switch(c->c.type) {
    case CATWEASEL_TYPE_MK1:
	release_region(c->c.iobase, 8);
	break;
    case CATWEASEL_TYPE_MK3:
	release_region(c->c.iobase, 256);
	break;
    }
}

static int find_free_controller_struct ()
{
    int i;

    for(i = 0; i < sizeof(controllers)/sizeof(controllers[0]); i++) {
	if(controllers[i].c.type == CATWEASEL_TYPE_NONE) {
	    return i;
	}
    }

    return -1;
}

static int add_isa_controller (int iobase)
{
    int err;
    int cn;
    struct cw_controller_struct *c;

    if(!iobase)
	return 0;

    if((cn = find_free_controller_struct()) < 0) {
	printk(KERN_NOTICE DEVICE_NAME ": More than two Catweasels are not supported.\n");
	return -ENODEV;
    }

    c = controllers + cn;
    c->c.type = CATWEASEL_TYPE_MK1;
    c->c.iobase = iobase;

    if(!request_region(c->c.iobase, 8, DEVICE_NAME)) {
	printk(KERN_ERR DEVICE_NAME ": io ports 0x%04x-0x%04x in use\n", c->c.iobase, c->c.iobase+7);
	return -EBUSY;
    }

    printk(KERN_NOTICE DEVICE_NAME ": cw%d is ISA Catweasel at 0x%04x\n", cn, iobase);

    if((err = init_controller(c))) {
	release_region(c->c.iobase, 8);
	return err;
    }

    return 0;
}

#if defined(CONFIG_PCI) && LINUX_VERSION_CODE >= 0x020400

static int __devinit cwfloppy_probe(struct pci_dev *pcidev, const struct pci_device_id *pciid)
{
    int cn;
    struct cw_controller_struct *c;
    int err;

    if((cn = find_free_controller_struct()) < 0) {
	printk(KERN_NOTICE DEVICE_NAME ": More than two Catweasels are not supported.\n");
	return -ENODEV;
    }

    if((err=pci_enable_device(pcidev)))
	return err;

    c = controllers + cn;
    c->c.type = CATWEASEL_TYPE_MK3;
    c->c.iobase = pci_resource_start(pcidev, 0);

    if(!request_region(c->c.iobase, 256, DEVICE_NAME)) {
	printk(KERN_ERR DEVICE_NAME ": io ports 0x%04x-0x%04x in use\n", c->c.iobase, c->c.iobase+255);
	return -EBUSY;
    }

    printk(KERN_NOTICE DEVICE_NAME ": cw%d is PCI Catweasel\n", cn);

#if 1
    outb(0xf1, c->c.iobase + 0x0);
    outb(0x00, c->c.iobase + 0x1);
    outb(0x00, c->c.iobase + 0x2);
    outb(0x00, c->c.iobase + 0x4);
    outb(0x00, c->c.iobase + 0x5);
    outb(0x00, c->c.iobase + 0x29);
    outb(0x00, c->c.iobase + 0x2b);
#endif

    if((err = init_controller(c))) {
	release_region(c->c.iobase, 256);
	return err;
    }

    return 0;
}

static struct pci_device_id id_table[] __devinitdata = {
    { 0xe159, 0x0001, 0x1212, 0x0002, 0, 0 },
    { 0, }
};

MODULE_DEVICE_TABLE(pci, id_table);

static struct pci_driver cwfloppy_driver = {
    name: "cwfloppy",
    id_table: id_table,
    probe: cwfloppy_probe
};

#endif /* CONFIG_PCI */

int init_module(void)
{
    int err;

    init_timer(&ms_timer);
    ms_timer.function = ms_timeout;
    cw_init_wq(&hw_wq);
    cw_init_wq(&fdc_wq);

    controllers[0].c.type = CATWEASEL_TYPE_NONE;
    controllers[1].c.type = CATWEASEL_TYPE_NONE;

#if defined(CONFIG_PCI) && LINUX_VERSION_CODE >= 0x020400
    /* First try to find a PCI Catweasel */
    if(pci_present()) {
	err = pci_module_init(&cwfloppy_driver);
	if(err && err != -ENODEV)
	    return err;
	if(controllers[0].c.type == CATWEASEL_TYPE_NONE) {
	    printk(KERN_NOTICE DEVICE_NAME ": No PCI Catweasels found.\n");
	}
    }
#endif /* CONFIG_PCI */

    /* Now check if the user has provided an IO base for ISA
       Catweasels */
    err = add_isa_controller(io[0]);
    if(err)
	goto err_isa_failed;
    err = add_isa_controller(io[1]);
    if(err)
	goto err_isa_failed;

    if(controllers[0].c.type == CATWEASEL_TYPE_NONE) {
	printk(KERN_NOTICE DEVICE_NAME ": No ISA Catweasels found. Use \"io=0xNNN\" with insmod\n");
	err = -ENODEV;
	goto err_isa_failed;
    }

    if(register_blkdev(MAJOR_NR, "cwfloppy", &cw_bops)) {
	printk(DEVICE_NAME ": unable to get major number %d.\n",
	       MAJOR_NR);
	return -EIO;
    }

    memset(&cwfloppy_sizes, 0, sizeof(cwfloppy_sizes));
    memset(&cwfloppy_blksizes, 0, sizeof(cwfloppy_blksizes));
    blk_size[MAJOR_NR] = cwfloppy_sizes;
    blksize_size[MAJOR_NR] = cwfloppy_blksizes;
#if LINUX_VERSION_CODE >= 0x020400
    blk_init_queue(BLK_DEFAULT_QUEUE(MAJOR_NR), DEVICE_REQUEST);
#else
    blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
#endif

    codec_msdos_init();

    if((errors.verbose = verbose))
	printk(DEVICE_NAME ": verbose error reporting enabled\n");
    if(readonly)
	printk(DEVICE_NAME ": writing disabled\n");

    return 0;

  err_isa_failed:
    free_controller(&controllers[1]);
    free_controller(&controllers[0]);
#if defined(CONFIG_PCI) && LINUX_VERSION_CODE >= 0x020400
    pci_unregister_driver(&cwfloppy_driver);
#endif
 
    return err;
}

void cleanup_module(void)
{
    free_controller(&controllers[1]);
    free_controller(&controllers[0]);
#if defined(CONFIG_PCI) && LINUX_VERSION_CODE >= 0x020400
    pci_unregister_driver(&cwfloppy_driver);
#endif

    if(unregister_blkdev(MAJOR_NR, "cwfloppy") != 0) {
	printk("cwfloppy: cleanup_module failed\n");
    }

#if LINUX_VERSION_CODE >= 0x020400
    blk_cleanup_queue(BLK_DEFAULT_QUEUE(MAJOR_NR));
#endif
}
